( $Id: colorname.frt,v 1.1 2010/12/17 12:11:20 leon Exp albert $ )
( Copyright{2005}: Albert van der Horst, HCC FIG Holland by GNU Public License)

\ ------------------- Auxiliary specific for colorforth ---------------------

REQUIRE BOUNDS

\ Throughout the following: colorname means a machine word with a
\ huffman encoded name a la colorforth.

DECIMAL
\ Offsets in these strings are the Huffman code for 4,5 and 7 length
\ chars, exclusive the Huffman determination prefix.
\ These are all the chars used in colorforth, in order of some internal
\ coding. One of the worst secrets, capitals are present too!
: ALL-CHARS " rtoeanismcylgfwdvpbhxuq0123456789j-k.z/;'!+@*,? RTOEANISMCYLGFWDVPBHXUQ0123456789J-K.Z/;:!+@*,?" ;

\ The addresses where start the 4,5,7 bit encoded chars.
: C4 ALL-CHARS DROP ;              \ Prefix 0
: C5 ALL-CHARS DROP 8 + ;          \ Prefix 01
: C7 ALL-CHARS DROP 16 + ;        \ Prefix 11
\ The strings that contain the 4,5,7 bit encoded chars.
: C4$ C4  8 ;              \ Prefix 0
: C5$ C5  8 ;          \ Prefix 01
: C7$ C7  32 ;        \ Prefix 11

\ For a COLORNAME, return a COLORNAME without its first char, and that CHAR.
: UNPACK
    DUP 0< 0= IF
        DUP   4 LSHIFT SWAP
        28 RSHIFT  C4 + C@
    ELSE 1 LSHIFT DUP 0< 0= IF
        DUP   4 LSHIFT SWAP
        28 RSHIFT  C5 + C@
    ELSE 1 LSHIFT
        DUP   5 LSHIFT SWAP
        27 RSHIFT  C7 + C@
    THEN THEN ;

\ Append COLORNAME to string at ``PAD'', return as (ascii) STRINGCONSTANT
: (CN>ASC)  BEGIN UNPACK PAD $C+ DUP 0= UNTIL DROP   PAD $@ ;

\ Copy COLORNAME to ``PAD'', return as (ascii) STRINGCONSTANT
\ To not upset the disassembler with conflicting names colorforthnames
\ are prepended with "X_".
: CN>ASC  "X_" PAD $!   (CN>ASC) ;

\ Print out a COLORNAME.
: .WORD   CN>ASC TYPE ;

\ Rotate left BITS over N , giving BITS.
: ROL    2DUP LSHIFT >R   32 SWAP - RSHIFT   R> OR ;

\ If the Huffman CODE has a LENGTH more than 32 , make CODE with at most LENGTH of 32
: CORRECT-ROL DUP 32 > IF 64 SWAP - ROL 32 THEN ;

\ Into PRECOLORNAME, NUMBER of valid bits, incorporate COLORCHARACTER HUFFMANLENGTH
\ return longer PRECOLORNAME, NUMBER of valid bits.
: CN-C+   ROT OVER + >R  \ R: new length
           SWAP >R   ROL R> OR   R>   CORRECT-ROL ;

\ Make PRECOLORNAME, NUMBER into COLORNAME.
\ Sometimes the ``NUMBER'' of bits is larger than 32,
\ if the last bits are zero.
: CN-ADJUST   32 MIN 32 SWAP - LSHIFT ;

\ \ Convert a COLORCHARACTER to ASCII.
\ : CC>
\     DUP 8 < IF 28 LSHIFT ELSE DUP 16 < IF 27 LSHIFT ELSE 25 LSHIFT THEN THEN
\     UNPACK SWAP DROP ;

\ Convert an ASCII to a colorcharacter and return THAT plus LENGTH
: >CCL   >R
        C4$ R@ $I DUP IF C4 - 4 ELSE DROP
        C5$ R@ $I DUP IF C5 - 16 OR 5 ELSE DROP
        C7$ R@ $I DUP IF C7 - 96 OR 7 ELSE DROP
            -1 ABORT" Not a colorforth character"
        THEN THEN THEN   RDROP ;

\ Convert a NAME to a COLORNAME.
: >CN    0 0 2SWAP BOUNDS DO I C@ >CCL CN-C+ LOOP CN-ADJUST ;

\ Convert a COLORCHARACTER to ASCII.
: C>   ALL-CHARS DROP + C@ ;

\ Convert an ASCII to a COLORCHARACTER
: >C   >R ALL-CHARS R> $I   DUP 0= ABORT" Not a colorforth character"
        ALL-CHARS DROP - ;

HEX
\ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

\ Comma in a word.
: AS-,   DUP AS-C, 8 RSHIFT DUP AS-C, 8 RSHIFT DUP AS-C, 8 RSHIFT AS-C, ;

\ For STRING: "it CONTAINS no upper case char".
: NO-UPPER
    BEGIN DUP WHILE 1- 2DUP + C@ &A &Z 1+ WITHIN IF 2DROP 0 EXIT THEN REPEAT
    2DROP -1 ;

\ For STRING: "its CONTAINS no lower case char".
: NO-LOWER
    BEGIN DUP WHILE 1- 2DUP + C@ &a &z 1+ WITHIN IF 2DROP 0 EXIT THEN REPEAT
    2DROP -1 ;

\ For STRING convert its first character to upper-case.
: FIRST>UP DROP DUP C@ 20 XOR SWAP C! ;

\ Convert ADDRESS LENGTH in place to lower case.
: $>LOWER BOUNDS DO I C@ &A &Z 1+ WITHIN IF I C@ 20 XOR I C! THEN LOOP ;

\ A defining word for a colorprefix "name" with INDEX.
\ Execute : a number is commaed in, during assembly.
: COLOR-PREFIX  CREATE ,   DOES>   @ SWAP >CN + AS-, ;

\ The action for green hex NUMBER. Assemble code that will make it green.
: (]$)    DUP F800,0000 AND IF 15 AS-, AS-, ELSE
    5 LSHIFT 16 + AS-, THEN ;

\ The words in the vocabulary ``EDIT-TIME'' understand the special denotations
\ in screens. It only contains prefix words.

VOCABULARY EDIT-TIME
CURRENT @    'EDIT-TIME >WID CURRENT !

\ A number prefixed with ``,'' is commaed in.
\ This is a cop-out and no longer used.
: ,    (WORD) EVALUATE AS-, ;   PREFIX

\ A number prefixed with ``:'' is a red colorword. (define)
: :    (WORD) >CN 3 + AS-, ;   PREFIX

\ A number prefixed with ``]'' is a green colorword. (compile)
: ]    (WORD) >CN 4 + AS-, ;   PREFIX

\ A number prefixed with ``['' is a yellow colorword.  (execute)
: [    (WORD) >CN 1 + AS-, ;   PREFIX

\ A number prefixed with ``^'' is a cyan colorword.  (postpone)
: ^    (WORD) >CN 7 + AS-, ;   PREFIX

\ A number prefixed with ``?'' is a violet colorword.  (variable)
: ?    (WORD) >CN 0C + AS-, ;   PREFIX

\ A number prefixed with ``.'' is a blue colorword.  (layout)
: .    (WORD) >CN 0E + AS-, ;   PREFIX

\ A number prefixed with ``('' is a white colorword.  (comment)
\ A loose ``('' is normal Forth comment.
\ Sort out three type of comment, lower, upper and capitalized.
: (
    IN @ C@ BL = IF &) PARSE 2DROP ELSE (WORD)
    2DUP NO-UPPER IF >CN  9 + AS-, ELSE
    2DUP NO-LOWER IF 2DUP $>LOWER >CN 0B + AS-, ELSE
    2DUP FIRST>UP >CN 0A + AS-, THEN THEN THEN
; PREFIX

\ A number prefixed with ``+'' extends the previous word.  (extend)
: +    (WORD) >CN 0 + AS-, ;   PREFIX

\ A number prefixed with ``#'' is a yellow decimal number.
: #    (WORD) DECIMAL EVALUATE 5 LSHIFT HEX 8 + AS-, ;   PREFIX

\ A number prefixed with ``$'' is a yellow hex number.
: $    (WORD) EVALUATE 5 LSHIFT HEX 18 + AS-, ;   PREFIX

\ A number prefixed with ``[#'' is a green decimal number.
: ]#    (WORD) DECIMAL EVALUATE 5 LSHIFT HEX 6 + AS-, ;   PREFIX

\ A number prefixed with ``]$'' is a green hex number.
: ]$    (WORD) EVALUATE    (]$) ;   PREFIX

\ A number prefixed with ``%$'' is a brown hex number.
: %$   (WORD) EVALUATE 12 AS-, AS-, ;   PREFIX

\ Remove ``EDIT-TIME'' wordlist from search, leave ADDRESS where assembly
\ starts.
: {   PREVIOUS  AS-HERE ;

CURRENT !

\ If the cell one cell off contains one of these we have to disassemble.
\ Green commaers.
10 BAG COMMA-BAG
    "1," >CN 4 + COMMA-BAG BAG+!
    "2," >CN 4 + COMMA-BAG BAG+!
    "3," >CN 4 + COMMA-BAG BAG+!
    ","  >CN 4 + COMMA-BAG BAG+!

\ For NUMBER of bytes to handle, assemble correct colorforth commaer.
: SELECT-COMMA   CELLS COMMA-BAG + @   AS-, ;

\ Comma in code from ADDRESS, restart the assembly process.
: }    AS-HERE OVER -   0 AS-,   SWAP @
  OVER 1 5 WITHIN 0= IF DUP . "not within expetaton" TYPE THEN
  OVER 4 + NEGATE AS-ALLOT   (]$)   SELECT-COMMA   EDIT-TIME ;
